@@ -1,9 +1,10 @@
-#!/usr/bin/perl
+#!/usr/bin/env perl
 
 # Aggregate a bunch of cover files into a Sereal "database".
 
 use strict;
 use warnings;
+use Getopt::Long;
 
 use Devel::QuickCover::Report;
 
@@ -12,9 +13,32 @@ my $QC_PATH       = '/tmp';
 my $QC_PREFIXES   = 'QC';      # Could be '(QC|CQ|XX)'
 my $QC_EXTENSIONS = 'txt';     # Could be '(txt|dat|db)'
 
+my $HELP_MSG =<<END;
+qc-aggregate.pl - Aggregate Devel::QuickCover cover files
+
+USE:
+
+    qc-aggregate.pl
+
+DESCRIPTION:
+
+qc-aggregate.pl aggregates all Devel::QuickCover files $QC_PATH/${QC_PREFIXES}_\*.$QC_EXTENSIONS
+into a Sereal database that can be converted into Devel::Cover output.
+The data is written to $QC_DATABASE. If this file exists, the new data
+is added to it.
+END
+
 exit main();
 
 sub main {
+    my %options = ();
+    GetOptions(\%options, 'help|h');
+
+    if (defined $options{help}) {
+        print $HELP_MSG;
+        return 0;
+    }
+
     my $report = Devel::QuickCover::Report->new;
 
     $report->load($QC_DATABASE)
